#include "demo.h"

/*
	Demo mapping data from witilt
	Based upon earlier Wavefront OBJ file viewing utility
	2005
	Mar   First version
         Ignore texture coordinates
	2007
   Aug   Added serial input from WiTilt
*/

OPTIONS options;
CAMERA  camera;
INTERFACESTATE interfacestate;

int nvertex = 0;
XYZ *vertices = NULL;

int nnormal = 0;
XYZ *normals = NULL;

int npolygon = 0;
IPOLYGON *polygons = NULL;

typedef struct {
	char name[64];
} GROUPS;
int ngroup = 0;
GROUPS *groups = NULL;

XYZ themin = {1e32,1e32,1e32},themax = {-1e32,-1e32,-1e32},themid;
double framerate = 0;

// Serial file reference
int fdes;
int usewitilt = FALSE;
#define AVERAGELEN 30
XYZ witilt_p[AVERAGELEN];

int main(int argc,char **argv)
{
	int i;
	unsigned int n,nlast = 0;
	double tstart;
	FILE *fptr;

   // The default camera attributes 
   CameraHome(0);
   camera.stereo       = NOSTEREO;
   camera.screenwidth  = 600;
   camera.screenheight = 600;

	// Default options 
	options.debug        = FALSE;
	options.fullscreen   = FALSE;
	options.windowdump   = FALSE;
	options.record       = FALSE;
	options.targetfps    = 30;
	options.rendermode   = WIREFRAME;
	options.showinfo     = FALSE;
	options.showhelp     = FALSE;
	options.showconstr   = FALSE;
	options.autorotate   = 0;
	options.witiltx      = TRUE;
   options.witilty      = FALSE;
   options.witiltz      = FALSE;

	// State of the input device, mouse in this case 
	interfacestate.button = -1;
	interfacestate.shift  = FALSE;
	interfacestate.mouseh = 0;
	interfacestate.mousev = 0;

	// Parse the command line arguments 
	if (argc < 2) 
		GiveUsage(argv[0]);
   for (i=1;i<argc-1;i++) {
      if (strcmp(argv[i],"-h") == 0) 
         GiveUsage(argv[0]);
      if (strcmp(argv[i],"-f") == 0)
         options.fullscreen = !options.fullscreen;
      if (strcmp(argv[i],"-s") == 0)
         camera.stereo = ACTIVESTEREO;
		if (strcmp(argv[i],"-ss") == 0) {
         camera.stereo = DUALSTEREO;
			camera.screenwidth *= 2;
		}
		if (strcmp(argv[i],"-rw") == 0)
         options.rendermode = WIREFRAME;
      if (strcmp(argv[i],"-rs") == 0)
         options.rendermode = SHADED;
      if (strcmp(argv[i],"-rp") == 0)
         options.rendermode = SPECULAR;
      if (strcmp(argv[i],"-a") == 0)
         options.autorotate = 1;
		if (strcmp(argv[i],"-d") == 0)
			options.debug = TRUE;
   }
	
	// Attempt to open the object file 
	if ((fptr = fopen(argv[argc-1],"r")) == NULL) {
		fprintf(stderr,"Failed to read file \"%s\"\n",argv[argc-1]);
		CleanExit();
	}
	ReadObj(fptr);
   fclose(fptr);
	themid = MidPoint(themin,themax);
	fprintf(stderr,"Read %d vertices\n",nvertex);
	fprintf(stderr,"Bounds: %g <= x <= %g\n",themin.x,themax.x);
   fprintf(stderr,"        %g <= y <= %g\n",themin.y,themax.y); 
   fprintf(stderr,"        %g <= z <= %g\n",themin.z,themax.z); 
   fprintf(stderr,"Read %d normals\n",nnormal);
   fprintf(stderr,"Read %d polygons\n",npolygon);
	fprintf(stderr,"Groups\n");
	for (i=0;i<ngroup;i++)
		fprintf(stderr,"   %s\n",groups[i].name);

   // Set things up and go 
   glutInit(&argc,argv);
   if (camera.stereo == ACTIVESTEREO)
		glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH | GLUT_STEREO);
	else
      glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);

	// Create the window and handlers 
   glutCreateWindow("OFF Viewer");
	glutReshapeWindow(camera.screenwidth,camera.screenheight);
   if (options.fullscreen)
      glutFullScreen();
   glutDisplayFunc(HandleDisplay);
	glutReshapeFunc(HandleReshape);
	CreateMenus();
   glutKeyboardFunc(HandleKeyboard);
   glutSpecialFunc(HandleSpecialKeyboard);
   glutMouseFunc(HandleMouse);
	glutMotionFunc(HandleMouseMotion);
	glutPassiveMotionFunc(HandlePassiveMotion);
	glutVisibilityFunc(HandleVisibility);      
	if (camera.stereo == DUALSTEREO)
		glutSetCursor(GLUT_CURSOR_NONE);
	glutIdleFunc(HandleIdle);
   CreateEnvironment();

   // Open port
   if ((fdes = OpenSerial("/dev/tty.BlueRadios-COM0-1")) < 0) {
      fprintf(stderr,"Port open failed\n");
      CleanExit();
   }  
   
   // Set port characteristics
   if (!SetupSerial(fdes,57600,8,1,'n')) {
      fprintf(stderr,"Port setup failed\n");
      CleanExit();
   }

	// Waiting for witilt reset, wait 10 seconds
	fprintf(stderr,"Reset WiTilt now ....\n");
	tstart = GetRunTime();
   for (;;) {
      if (ReadSerialChar(fdes,&n)) {
         n &= 0x000000ff; // Just look at low order byte
         if (n == 0xE0 && nlast == 0xE0) {
            sleep(2);
            SendSerialString(fdes,"1");
				usewitilt = TRUE;
				break;
         }
         if (n >= ' ' && n <= '~')
            fprintf(stderr,"%c",n);
         else if (n == 9) // tab
            fprintf(stderr," ");
         else if (n == '\n')
            fprintf(stderr,"\n");
         else if (n == '\r')
            ;
         else
            fprintf(stderr,"%0x\n",n);
			nlast = n;
      }
      if (GetRunTime() - tstart > 10)
         break;
   }
	if (!usewitilt)
		fprintf(stderr,"Didn't get a WiTilt reset in 10 seconds!\n");
	else
		fprintf(stderr,"Hold in zero position\n");

	// Ready to go! 
	CameraHome(0);
   glutMainLoop();
   return(0);
}

/*
   Form the menu structure
	Associate the menus with the right mouse button
*/
void CreateMenus(void)
{
	int mainmenu,cameramenu,rendermenu,optionsmenu;

   cameramenu = glutCreateMenu(HandleCameraMenu);
   glutAddMenuEntry("Front  [F1]",1);
   glutAddMenuEntry("Back   [F2]",2);
   glutAddMenuEntry("Left   [F3]",3);
   glutAddMenuEntry("Right  [F4]",4);
   glutAddMenuEntry("Top    [F5]",5);
   glutAddMenuEntry("Bottom [F6]",6);

   rendermenu = glutCreateMenu(HandleRenderMenu);
	glutAddMenuEntry("Points [1]",POINTS);
   glutAddMenuEntry("Wireframe [2]",WIREFRAME);
   glutAddMenuEntry("Shaded [3]",SHADED);
	glutAddMenuEntry("Specular [4]",SPECULAR);

   optionsmenu = glutCreateMenu(HandleOptionsMenu);
   glutAddMenuEntry("Interface information [h]",1);
   glutAddMenuEntry("Construction lines [c]",2);
   glutAddMenuEntry("Information [i]",3);

   mainmenu = glutCreateMenu(HandleMainMenu);
   glutAddSubMenu("Camera",cameramenu);
	glutAddSubMenu("Render",rendermenu);
	glutAddSubMenu("Options",optionsmenu);
   glutAddMenuEntry("Quit [ESC]",10);

   glutAttachMenu(GLUT_RIGHT_BUTTON);
}

/*
   This is where global OpenGL settings are made.
	Since this is not designed to demonstrate how to render
	objects in OpenGL, it's pretty boring. 
*/
void CreateEnvironment(void)
{
	/* Miscellaneous settings */
	glDisable(GL_CULL_FACE);
   glEnable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
   glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glPixelStorei(GL_UNPACK_ALIGNMENT,1);

	/* These are hardware dependent */
	glDisable(GL_DITHER);
   glDisable(GL_LINE_SMOOTH);
   glDisable(GL_POINT_SMOOTH);
   glDisable(GL_POLYGON_SMOOTH);
   glHint(GL_POINT_SMOOTH_HINT,GL_FASTEST);
   glHint(GL_LINE_SMOOTH_HINT,GL_FASTEST);
   glHint(GL_POLYGON_SMOOTH_HINT,GL_FASTEST);
	glHint(GL_FOG_HINT,GL_FASTEST);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_FASTEST);

	/* Default geometry rendering */
   glLineWidth(1.0);
   glPointSize(1.0);
   glFrontFace(GL_CCW);
   glShadeModel(GL_SMOOTH);
   glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);

   /* Default meterials */
	glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);

   /* Turn off all the lights */
   glDisable(GL_LIGHT0);
   glDisable(GL_LIGHT1);
   glDisable(GL_LIGHT2);
   glDisable(GL_LIGHT3);
   glDisable(GL_LIGHT4);
   glDisable(GL_LIGHT5);
   glDisable(GL_LIGHT6);
   glDisable(GL_LIGHT7);
   glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,GL_FALSE);
   glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
	glClearColor(1.0,1.0,1.0,0.0);
}

/*
   This is the basic display callback routine
	This is where the real projection work for stereo
	is done, note the viewer is assumed to be located
	centrally to the screen and looking along a vector
	perpendicular to the screen.
*/
void HandleDisplay(void)
{
	XYZ r;
   double ratio,radians,wd2,ndfl;
   double left,right,top,bottom;
	double dt;
	static int framecounter = 0;
	static double tstart = -1;

	if (tstart < 0)
		tstart = GetRunTime();

   /* Misc stuff needed for the frustum */
   ratio   = camera.screenwidth / (double)camera.screenheight;
	if (camera.stereo == DUALSTEREO)
		ratio /= 2;
   radians = DTOR * camera.aperture / 2;
   wd2     = camera.near * tan(radians);
   ndfl    = camera.near / camera.focallength;
   top     =   wd2;
   bottom  = - wd2;

	/* Determine the right eye vector */
   CROSSPROD(camera.vd,camera.vu,r);
   Normalise(&r);
   r.x *= camera.eyesep / 2.0;
   r.y *= camera.eyesep / 2.0;
   r.z *= camera.eyesep / 2.0;

	if (camera.stereo == ACTIVESTEREO || camera.stereo == DUALSTEREO) {

      if (camera.stereo == DUALSTEREO) {
         glDrawBuffer(GL_BACK);
         glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
      }

   	glMatrixMode(GL_PROJECTION);
   	glLoadIdentity();
      left  = - ratio * wd2 - 0.5 * camera.eyesep * ndfl;
      right =   ratio * wd2 - 0.5 * camera.eyesep * ndfl;
      glFrustum(left,right,bottom,top,camera.near,camera.far);
		if (camera.stereo == DUALSTEREO)
			glViewport(camera.screenwidth/2,0,camera.screenwidth/2,camera.screenheight);
		else
			glViewport(0,0,camera.screenwidth,camera.screenheight);
      glMatrixMode(GL_MODELVIEW);
      glDrawBuffer(GL_BACK_RIGHT);
      if (camera.stereo == ACTIVESTEREO)
         glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
      glLoadIdentity();
      gluLookAt(camera.vp.x + r.x,camera.vp.y + r.y,camera.vp.z + r.z,
                camera.vp.x + r.x + camera.vd.x,
                camera.vp.y + r.y + camera.vd.y,
                camera.vp.z + r.z + camera.vd.z,
                camera.vu.x,camera.vu.y,camera.vu.z);
      CreateGeometry();

      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      left  = - ratio * wd2 + 0.5 * camera.eyesep * ndfl;
      right =   ratio * wd2 + 0.5 * camera.eyesep * ndfl;
      glFrustum(left,right,bottom,top,camera.near,camera.far);
      if (camera.stereo == DUALSTEREO)
         glViewport(0,0,camera.screenwidth/2,camera.screenheight);
      else
         glViewport(0,0,camera.screenwidth,camera.screenheight);
      glMatrixMode(GL_MODELVIEW);
      glDrawBuffer(GL_BACK_LEFT);
      if (camera.stereo == ACTIVESTEREO)
         glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
      glLoadIdentity();
      gluLookAt(camera.vp.x - r.x,camera.vp.y - r.y,camera.vp.z - r.z,
                camera.vp.x - r.x + camera.vd.x,
                camera.vp.y - r.y + camera.vd.y,
                camera.vp.z - r.z + camera.vd.z,
                camera.vu.x,camera.vu.y,camera.vu.z);
      CreateGeometry();

	} else {

      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
		glViewport(0,0,camera.screenwidth,camera.screenheight); 
      left  = - ratio * wd2;
      right =   ratio * wd2;
      glFrustum(left,right,bottom,top,camera.near,camera.far);
      glMatrixMode(GL_MODELVIEW);
      glDrawBuffer(GL_BACK);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
      glLoadIdentity();
      gluLookAt(camera.vp.x,camera.vp.y,camera.vp.z,
                camera.vp.x + camera.vd.x,
                camera.vp.y + camera.vd.y,
                camera.vp.z + camera.vd.z,
                camera.vu.x,camera.vu.y,camera.vu.z);
   	CreateGeometry();
	}

	// Dump the window in compressed TGA format 
	if (options.record || options.windowdump) {
		if (camera.stereo == ACTIVESTEREO)
			WindowDump("",camera.screenwidth,camera.screenheight,TRUE,12);
		else
			WindowDump("",camera.screenwidth,camera.screenheight,FALSE,12);
		options.windowdump = FALSE;
	}

   // Swap buffers 
   glutSwapBuffers();

   // Compute the instantaneous frame rate 
   dt = GetRunTime() - tstart;
   if (dt > 0.5) {
      framerate = framecounter / dt;
      tstart = GetRunTime();
      framecounter = 0;
   }
   framecounter++;

	if (options.autorotate != 0)
		RotateCamera(1.0,0.0,0.0,options.autorotate/5.0);
}

/*
	Sort out rendering and lighting modes 
*/
void MakeLighting(void)
{
   GLfloat p[4];
   GLfloat white[4] = {1.0,1.0,1.0,1.0};
	GLfloat midgrey[4] = {0.2,0.2,0.2,1.0};
	GLfloat black[4] = {0.0,0.0,0.0,1.0};
	GLfloat shiny[1] = {100.0};

	/* Create lights */
	p[0] = themid.x + (themax.x - themin.x);
	p[1] = themid.y + (themax.y - themin.y);
	p[2] = themid.z + (themax.z - themin.z);
	p[3] = 1.0;
	glLightfv(GL_LIGHT0,GL_POSITION,p);
   glLightfv(GL_LIGHT0,GL_DIFFUSE,white);
	switch (options.rendermode) {
	case POINTS:
	case WIREFRAME:
	case SHADED:
		glLightfv(GL_LIGHT0,GL_SPECULAR,black);
		break;
	case SPECULAR:
		glLightfv(GL_LIGHT0,GL_SPECULAR,white);
		break;
	}
	glLightfv(GL_LIGHT0,GL_AMBIENT,black);
	glEnable(GL_LIGHT0);

   p[0] = camera.vp.x + camera.vu.x;
   p[1] = camera.vp.y + camera.vu.y;
   p[2] = camera.vp.z + camera.vu.z;
   p[3] = 1.0;
   glLightfv(GL_LIGHT0,GL_POSITION,p);
   glLightfv(GL_LIGHT0,GL_DIFFUSE,white);
   switch (options.rendermode) {
	case POINTS:
   case WIREFRAME:
   case SHADED:
      glLightfv(GL_LIGHT1,GL_SPECULAR,black);
      break;
   case SPECULAR:
      glLightfv(GL_LIGHT1,GL_SPECULAR,white);
      break;
   }
   glLightfv(GL_LIGHT0,GL_AMBIENT,black);

   /* Rendering mode specific settings */
   switch (options.rendermode) {
	case POINTS:
   case WIREFRAME:
      glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
      glLightModelfv(GL_LIGHT_MODEL_AMBIENT,white);
      glDisable(GL_LIGHTING);
      break;
   case SHADED:
	case SPECULAR:
      glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
      glLightModelfv(GL_LIGHT_MODEL_AMBIENT,midgrey);
      glEnable(GL_LIGHTING);
      break;
   }

	/* Materials */
   switch (options.rendermode) {
	case POINTS:
   case WIREFRAME:
      glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,white);
      glMaterialfv(GL_FRONT_AND_BACK,GL_SHININESS,shiny);
		glMaterialfv(GL_FRONT_AND_BACK,GL_EMISSION,black);
      break;
   case SHADED:
	case SPECULAR:
      glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,white);
      glMaterialfv(GL_FRONT_AND_BACK,GL_SHININESS,shiny);
		glMaterialfv(GL_FRONT_AND_BACK,GL_EMISSION,black);
		break;
	}
}

/*
	Create the geometry
*/
void CreateGeometry(void)
{
	int i,j;
	XYZ p[3],n[3];
	XYZ theta;

	MakeLighting();
	glColor3f(0.5,0.8,0.5);

	theta = GetWitiltPoint();
	if (options.witiltx) {
		glRotatef(-theta.x,1,0,0);
      fprintf(stderr,"thetax: %g\n",theta.x);
	}
	if (options.witilty) {
   	glRotatef(-theta.y,0,0,1);
      fprintf(stderr,"thetay: %g\n",theta.y);
	}
	if (options.witiltz) {
   	glRotatef(theta.z,0,1,0);
		fprintf(stderr,"thetaz: %g\n",theta.z);
	}

	glRotatef(90,0,1,0);
	if (options.rendermode == POINTS) {
		glBegin(GL_POINTS);
		for (i=0;i<nvertex;i++) 
			glVertex3f(vertices[i].x,vertices[i].y,vertices[i].z);
		glEnd();
	} else {
		glBegin(GL_TRIANGLES);
		for (i=0;i<npolygon;i++) {
     	   p[0] = vertices[polygons[i].p1];
     	   p[1] = vertices[polygons[i].p2];
     	   p[2] = vertices[polygons[i].p3];
         n[0] = normals[polygons[i].n1];
         n[1] = normals[polygons[i].n2];
         n[2] = normals[polygons[i].n3];
			for (j=0;j<3;j++) {
     	   	glNormal3f(n[j].x,n[j].y,n[j].z);
     	   	glVertex3f(p[j].x,p[j].y,p[j].z);
			}
		}
     	glEnd();
	}

	DrawInfoHelp();
}

/*
   Deal with plain key strokes
*/
void HandleKeyboard(unsigned char key,int x, int y)
{
   if (glutGetModifiers() == GLUT_ACTIVE_SHIFT)
      interfacestate.shift = TRUE;
   else
      interfacestate.shift = FALSE;

	/* Handle keyboard commands */
   switch (key) {
   case ESC: 		
   case 'Q':
   case 'q': 
		CleanExit();
		break;
   case '1':
      options.rendermode = POINTS;
      break;
	case '2':
   	options.rendermode = WIREFRAME;
		break;
   case '3':
      options.rendermode = SHADED;
      break;
   case '4':
      options.rendermode = SPECULAR;
      break;
	case 'a':
		options.autorotate--;
		break;
	case 'A':
		options.autorotate++;
		break;
   case 'i':
   case 'I':
      options.showinfo = !options.showinfo;
      break;
	case 'h':
	case 'H':
		options.showhelp = !options.showhelp;
      break;
   case 'c':
   case 'C':
      options.showconstr = !options.showconstr;
      break;
   case '[':
      RotateCamera(0.0,0.0,-1.0,1.0);
      break;
   case ']':
      RotateCamera(0.0,0.0,1.0,1.0);
      break;
   case 'w':
	case 'W':
		options.windowdump = !options.windowdump;
      break;
   case 'r':
   case 'R':
      options.record = !options.record;
      break;
   case '<':
   case ',':
		TranslateCamera(0.0,0.0,-camera.focallength/50);
		break;
   case '>':
   case '.':
		TranslateCamera(0.0,0.0,+camera.focallength/50);
      break;
	case '-':
	case '_':
		if (camera.aperture < 90)
			camera.aperture *= 1.05;
		break;
	case '+':
   case '=':
		if (camera.aperture > 5)
			camera.aperture /= 1.05;
      break;
	case '(':
	case '9':
		camera.focallength *= 1.05;
		break;
	case ')':
	case '0':
		camera.focallength /= 1.05;
		break;
	case 'x':
	case 'X':
		options.witiltx = !options.witiltx;
		break;
   case 'y':
   case 'Y':
      options.witilty = !options.witilty;
      break;
   case 'z':
   case 'Z':
      options.witiltz = !options.witiltz;
      break;
	}
}

/*
   Deal with special key strokes
*/
void HandleSpecialKeyboard(int key,int x, int y)
{
   if (glutGetModifiers() == GLUT_ACTIVE_SHIFT)
      interfacestate.shift = TRUE;
   else
      interfacestate.shift = FALSE;

   switch (key) {
   case GLUT_KEY_LEFT:  
		if (interfacestate.shift)
			TranslateCamera(-camera.focallength/200,0.0,0.0);
		else
			RotateCamera(-1.0,0.0,0.0,1.0); 
		break;
   case GLUT_KEY_RIGHT: 
      if (interfacestate.shift)
         TranslateCamera(camera.focallength/200,0.0,0.0);
      else
			RotateCamera(1.0,0.0,0.0,1.0);  
		break;
   case GLUT_KEY_UP:    
      if (interfacestate.shift)
         TranslateCamera(0.0,camera.focallength/200,0.0);
      else
			RotateCamera(0.0,1.0,0.0,1.0);  
		break;
   case GLUT_KEY_DOWN:  
      if (interfacestate.shift)
         TranslateCamera(0.0,-camera.focallength/200,0.0);
      else
			RotateCamera(0.0,-1.0,0.0,1.0); 
		break;
   case GLUT_KEY_F1:
      CameraHome(1);
      break;
   case GLUT_KEY_F2:
      CameraHome(2);
      break;
   case GLUT_KEY_F3:
      CameraHome(3);
      break;
   case GLUT_KEY_F4:
      CameraHome(4);
      break;
   case GLUT_KEY_F5:
      CameraHome(5);
      break;
   case GLUT_KEY_F6:
      CameraHome(6);
      break;
   }
}

/*
   Rotate (ix,iy) or roll (iz) the camera about the focal point
   ix,iy,iz are flags, 0 do nothing, +- 1 rotates in opposite directions
   Correctly updating all camera attributes
*/
void RotateCamera(double ix,double iy,double iz,double dtheta)
{
   XYZ vp,vu,vd;
   XYZ right;
   XYZ newvp,newr;
   double radius,dd,radians;
   double dx,dy,dz;

   vu = camera.vu;
   Normalise(&vu);
   vp = camera.vp;
   vd = camera.vd;
   Normalise(&vd);
   CROSSPROD(vd,vu,right);
   Normalise(&right);
   radians = dtheta * PI / 180.0;

   /* Handle the roll */
   if (ABS(iz) > 0.0001) {
      camera.vu.x += iz * right.x * radians;
      camera.vu.y += iz * right.y * radians;
      camera.vu.z += iz * right.z * radians;
      Normalise(&camera.vu);
      return;
   }
 
   /* Distance from the rotate point */
   dx = camera.vp.x - camera.pr.x;
   dy = camera.vp.y - camera.pr.y;
   dz = camera.vp.z - camera.pr.z;
   radius = sqrt(dx*dx + dy*dy + dz*dz);
 
   /* Determine the new view point */
   dd = radius * radians;
   newvp.x = vp.x + dd * ix * right.x + dd * iy * vu.x - camera.pr.x;
   newvp.y = vp.y + dd * ix * right.y + dd * iy * vu.y - camera.pr.y;
   newvp.z = vp.z + dd * ix * right.z + dd * iy * vu.z - camera.pr.z;
   Normalise(&newvp);
   camera.vp.x = camera.pr.x + radius * newvp.x;
   camera.vp.y = camera.pr.y + radius * newvp.y;
   camera.vp.z = camera.pr.z + radius * newvp.z;
 
   /* Determine the new right vector */
   newr.x = camera.vp.x + right.x - camera.pr.x;
   newr.y = camera.vp.y + right.y - camera.pr.y;
   newr.z = camera.vp.z + right.z - camera.pr.z;
   Normalise(&newr);
   newr.x = camera.pr.x + radius * newr.x - camera.vp.x;
   newr.y = camera.pr.y + radius * newr.y - camera.vp.y;
   newr.z = camera.pr.z + radius * newr.z - camera.vp.z;
   Normalise(&newr);
 
   camera.vd.x = camera.pr.x - camera.vp.x;
   camera.vd.y = camera.pr.y - camera.vp.y;
   camera.vd.z = camera.pr.z - camera.vp.z;
   Normalise(&camera.vd);
 
   /* Determine the new up vector */
   CROSSPROD(newr,camera.vd,camera.vu);
   Normalise(&camera.vu);
}

/*
   Translate (pan) the camera view point
*/
void TranslateCamera(double ix,double iy,double iz)
{
   XYZ right;

   CROSSPROD(camera.vd,camera.vu,right);
   Normalise(&right);

   camera.vp.x += iy*camera.vu.x + ix*right.x + iz*camera.vd.x;
   camera.vp.y += iy*camera.vu.y + ix*right.y + iz*camera.vd.y;
   camera.vp.z += iy*camera.vu.z + ix*right.z + iz*camera.vd.z;

   camera.pr.x += iy*camera.vu.x + ix*right.x + iz*camera.vd.x;
   camera.pr.y += iy*camera.vu.y + ix*right.y + iz*camera.vd.y;
   camera.pr.z += iy*camera.vu.z + ix*right.z + iz*camera.vd.z;
}

/*
   Handle mouse events.
	Do nothing special, just remember the state.
*/
void HandleMouse(int button,int state,int x,int y)
{
   if (glutGetModifiers() == GLUT_ACTIVE_SHIFT)
      interfacestate.shift = TRUE;
   else
      interfacestate.shift = FALSE;

   if (state == GLUT_DOWN) {
      if (button == GLUT_LEFT_BUTTON) {
			interfacestate.button = GLUT_LEFT_BUTTON;
      } else if (button == GLUT_MIDDLE_BUTTON) {
			interfacestate.button = GLUT_MIDDLE_BUTTON;
      } else {
			/* Assigned to menus */
		}
   }
   interfacestate.mouseh = x;
   interfacestate.mousev = y;
}

/*
   Handle mouse motion
*/
void HandleMouseMotion(int x,int y)
{
   double dx,dy;

   dx = x - interfacestate.mouseh;
   dy = y - interfacestate.mousev;

   if (interfacestate.button == GLUT_LEFT_BUTTON) {
		if (interfacestate.shift)
			TranslateCamera(-camera.focallength*dx/100,camera.focallength*dy/100,0.0);
		else
      	RotateCamera(-dx,dy,0.0,0.5);
   } else if (interfacestate.button == GLUT_MIDDLE_BUTTON) {
      if (interfacestate.shift)
         TranslateCamera(0.0,0.0,camera.focallength*dx/100);
      else
      	RotateCamera(0.0,0.0,dx,0.5);
	}

   interfacestate.mouseh = x;
   interfacestate.mousev = y;
}

/*
	What to do on moving mouse, the button isn't down.
*/
void HandlePassiveMotion(int x,int y)
{
   interfacestate.mouseh = x;
   interfacestate.mousev = y;
}


/*
   Handle the main menu
*/
void HandleMainMenu(int whichone)
{
   switch (whichone) {
   case 10: 
		CleanExit();
		break;
   }
}

/*
   Handle the camera menu
*/
void HandleCameraMenu(int whichone)
{
	CameraHome(whichone);
}

/*
   Handle the options menu
*/
void HandleOptionsMenu(int whichone)
{
	switch (whichone) {
	case 1:
		options.showhelp = !options.showhelp;
		break;
	case 2:
		options.showconstr = !options.showconstr;
		break;
	case 3:
		options.showinfo = !options.showinfo;
		break;
	}
}

/*
   Handle the render menu
*/
void HandleRenderMenu(int whichone)
{
   options.rendermode = whichone;
}

/*
   How to handle visibility
	If the window is not visible then don't draw it,
	allows layered windows without a processor load blowout.
*/
void HandleVisibility(int visible)
{
   if (visible == GLUT_VISIBLE)
      glutIdleFunc(HandleIdle);
	else
		glutIdleFunc(NULL);
}

/*
   What to do on an idle event.
	Maintain a constant frame rate.
*/
void HandleIdle(void)
{
	int i;
	unsigned int n,nlast=0;
   static double tstart = -1;
	static XYZ p;
   double tstop;

	static int buffptr = 0;
	static char buffer[256] = "";

	// Are we using the witilt?
	if (usewitilt) {
      if (ReadSerialChar(fdes,&n)) {
         n &= 0x000000ff; // Just look at low order byte
         if (n == 0xE0 && nlast == 0xE0) {
				fprintf(stderr,"Resetting the WiTilt\n");
            sleep(2);
            SendSerialString(fdes,"1");
         }
         if (n > ' ' && n <= '~') {
				buffer[buffptr] = n;
				buffer[buffptr+1] = '\0';
				buffptr++;
         } else if (n == '\r') { // process the string
				for (i=0;i<strlen(buffer);i++) {
					if (buffer[i] < '0' || buffer[i] > '9')
						buffer[i] = ' ';
				}
				if (sscanf(buffer,"%lf %lf %lf",&p.x,&p.y,&p.z) == 3) {
					AddWitiltPoint(p);
					//fprintf(stderr,"%g %g %g\n",p.x,p.y,p.z);
				} else {
					fprintf(stderr,"Got bad point\n");
				}
				buffptr = 0;
				buffer[0] = '\0';
			}
         nlast = n;
      }
	}

   if (tstart < 0)
      tstart = GetRunTime();
   tstop = GetRunTime();
   if (tstop - tstart > 1.0/options.targetfps) {
      // Nasty hack to solve Mac OS-X cursor problem
      if (camera.stereo == DUALSTEREO) {
         glutSetCursor(GLUT_CURSOR_CROSSHAIR);
         glutSetCursor(GLUT_CURSOR_NONE);
      }
      glutPostRedisplay();
      tstart = tstop;
   }
}

/*
   Handle a window reshape/resize
*/
void HandleReshape(int w,int h)
{
   camera.screenwidth = w;
   camera.screenheight = h;
}

/*
   Display the program usage information
*/
void GiveUsage(char *cmd)
{
   fprintf(stderr,"Usage: %s [Options] obj_file_name\n",cmd);
	fprintf(stderr,"Options\n");
   fprintf(stderr,"   -h   this text\n");
	fprintf(stderr,"   -f   full screen\n");
	fprintf(stderr,"   -s   active stereo\n");
	fprintf(stderr,"  -ss   dual screen stereo\n");
	fprintf(stderr,"   -a   auto rotate\n");
	fprintf(stderr,"  -rw   wireframe\n");
   fprintf(stderr,"  -rs   flat shaded\n");
   fprintf(stderr,"  -rp   specular shaded\n");
   exit(-1);
}

/*
	Move the camera to the home position 
*/
void CameraHome(int position)
{
	XYZ up = {0.0,1.0,0.0},x = {1.0,0.0,0.0};

	camera.aperture = 60;
	camera.focallength = VectorLength(themin,themax);
   camera.eyesep = camera.focallength / 25;
   camera.near = camera.focallength / 10;
   camera.far = camera.focallength * 10;

	//options.autorotate = 0;

	switch (position) {
	case 0:
		camera.vp.x = themid.x - camera.focallength;
		camera.vp.y = themid.y; 
		camera.vp.z = themid.z;
   	camera.vd.x = 1;
      camera.vd.y = 0;
      camera.vd.z = 0;
   	camera.vu = up;
		break;
	case 1: /* Front */
      camera.vp.x = themid.x - camera.focallength;
      camera.vp.y = themid.y;
      camera.vp.z = themid.z;
      camera.vd.x = 1;
      camera.vd.y = 0;
      camera.vd.z = 0;
		camera.vu = up;
		break;
	case 2: /* Back */
      camera.vp.x = themid.x + camera.focallength;
      camera.vp.y = themid.y;
      camera.vp.z = themid.z;
      camera.vd.x = -1;
      camera.vd.y = 0;
      camera.vd.z = 0;
		camera.vu = up;
		break;
	case 3: /* Left */
      camera.vp.x = themid.x;
      camera.vp.y = themid.y + camera.focallength;
      camera.vp.z = themid.z;
      camera.vd.x = 0;
      camera.vd.y = -1;
      camera.vd.z = 0;
		camera.vu = up;
		break;
	case 4: /* Right */
      camera.vp.x = themid.x;
      camera.vp.y = themid.y - camera.focallength;
      camera.vp.z = themid.z;
      camera.vd.x = 0;
      camera.vd.y = 1;
      camera.vd.z = 0;
		camera.vu = up;
		break;
	case 5: /* Top */
      camera.vp.x = themid.x;
      camera.vp.y = themid.y;
      camera.vp.z = themid.z + camera.focallength;
      camera.vd.x = 0;
      camera.vd.y = 0;
      camera.vd.z = -1;
		camera.vu = x;
		break;
	case 6: /* Bottom */
      camera.vp.x = themid.x;
      camera.vp.y = themid.y; 
      camera.vp.z = themid.z - camera.focallength;
      camera.vd.x = 0;
      camera.vd.y = 0;
      camera.vd.z = 1;
		camera.vu = x;
		break;
	}

	camera.pr = themid;

	Normalise(&camera.vd);
	Normalise(&camera.vu);
}

/*
   Draw various bits of information 
*/
void DrawInfoHelp(void)
{
   char s[256];
   int y = 6;
	double len;
	XYZ p;

   if (options.showinfo) {
		glDisable(GL_LIGHTING);
		glColor3f(1.0,1.0,1.0);

   	sprintf(s,"Vertices: %d, Polygons: %d",nvertex,npolygon);
   	DrawGLText(10,y+=14,s);

   	sprintf(s,"Frame rate: %.1f",framerate);
   	DrawGLText(10,y+=14,s);

		sprintf(s,"        %g <= z <= %g",themin.z,themax.z);
		DrawGLText(10,y+=14,s);
      sprintf(s,"        %g <= y <= %g",themin.y,themax.y);
      DrawGLText(10,y+=14,s);
      sprintf(s,"Bounds: %g <= x <= %g",themin.x,themax.x);
      DrawGLText(10,y+=14,s);
	}

	if (options.showconstr) {

		/* Draw the rotation point */
		len = camera.focallength / 20;
		glColor3f(0.5,1.0,0.5);
   	glBegin(GL_LINES);
   	glVertex3f(camera.pr.x-len,camera.pr.y,camera.pr.z); /* x */
   	glVertex3f(camera.pr.x+len,camera.pr.y,camera.pr.z);
   	glVertex3f(camera.pr.x,camera.pr.y-len,camera.pr.z); /* y */
   	glVertex3f(camera.pr.x,camera.pr.y+len,camera.pr.z);
   	glVertex3f(camera.pr.x,camera.pr.y,camera.pr.z-len); /* z */
   	glVertex3f(camera.pr.x,camera.pr.y,camera.pr.z+len);
   	glEnd();
	
      /* Draw the focal point */
      len = camera.focallength / 25;
		p.x = camera.vp.x + camera.focallength * camera.vd.x;
      p.y = camera.vp.y + camera.focallength * camera.vd.y; 
      p.z = camera.vp.z + camera.focallength * camera.vd.z; 
      glColor3f(1.0,0.0,0.0);
      glBegin(GL_LINES);
      glVertex3f(p.x-len,p.y,p.z); /* x */
      glVertex3f(p.x+len,p.y,p.z);
      glVertex3f(p.x,p.y-len,p.z); /* y */
      glVertex3f(p.x,p.y+len,p.z);
      glVertex3f(p.x,p.y,p.z-len); /* z */
      glVertex3f(p.x,p.y,p.z+len);
      glEnd();

		/* Draw the bounding box */
		glColor3f(1.0,0.5,0.5);
		CreateBox(themin,themax,0);

		/* Draw the axes */
		glBegin(GL_LINES);
		glColor3f(1.0,0.0,0.0);
		glVertex3f(0.0,0.0,0.0);
		glVertex3f(len,0.0,0.0);
   	glColor3f(0.0,1.0,0.0);
   	glVertex3f(0.0,0.0,0.0);
   	glVertex3f(0.0,len,0.0);
   	glColor3f(0.0,0.0,1.0);
   	glVertex3f(0.0,0.0,0.0);
   	glVertex3f(0.0,0.0,len);
		glEnd();
	}

	if (options.showhelp) {
		glColor3f(1.0,1.0,1.0);
		DrawGLText(10,camera.screenheight-20,interfacestring);
	}
}

/*
	Read a wavefront OBJ file, return status
*/
int ReadObj(FILE *fptr)
{
	int i,found;
	int igroup = 0; // Default
	int linecount = 0;
	int iv1,it1,in1,iv2,it2,in2,iv3,it3,in3;
	char aline[1024];
	char cmd[1024];
	XYZ p;

	ngroup = 1;
	groups = realloc(groups,sizeof(GROUPS));
	strcpy(groups[0].name,"default");
	
	while (ReadLine(fptr,aline,1000)) {
		linecount++;
		sscanf(aline,"%s",cmd);

		// Vertices
		if (strcmp(cmd,"v") == 0) {
			vertices = realloc(vertices,(nvertex+1)*sizeof(XYZ));
			if (sscanf(aline,"%s %lf %lf %lf",cmd,&p.x,&p.y,&p.z) != 4) {
				fprintf(stderr,"Unexpected line at %d\n",linecount);
				continue;
			}
			vertices[nvertex] = p;
			nvertex++;
			MinMaxXYZ(p,&themin,&themax);
		}

		// Texture coordinates
      if (strcmp(cmd,"vt") == 0) {
         
      }

		// Normals
      if (strcmp(cmd,"vn") == 0) {
         normals = realloc(normals,(nnormal+1)*sizeof(XYZ));
         if (sscanf(aline,"%s %lf %lf %lf",cmd,&p.x,&p.y,&p.z) != 4) {
            fprintf(stderr,"Unexpected line at %d\n",linecount);
            continue;
         }
         normals[nnormal] = p;
         nnormal++;
      }

		// Groups
      if (strcmp(cmd,"g") == 0) {
			strcpy(cmd,&(aline[2]));
			found = FALSE;
			for (i=0;i<ngroup;i++) {
				if (strcmp(groups[i].name,cmd) == 0) {
					igroup = i;
					break;
				}
			}
			if (!found) {
				groups = realloc(groups,(ngroup+1)*sizeof(GROUPS));
				strcpy(groups[ngroup].name,cmd);
				igroup = ngroup;
				ngroup++;
			}
      }

      // Faces
      if (strcmp(cmd,"f") == 0) {
			for (i=0;i<(int)strlen(aline);i++)
				if (aline[i] == '/')
					aline[i] = ' ';		
         polygons = realloc(polygons,(npolygon+1)*sizeof(IPOLYGON));
         if (sscanf(aline,"%s %d %d %d %d %d %d %d %d %d",
				cmd,&iv1,&it1,&in1,&iv2,&it2,&in2,&iv3,&it3,&in3) != 10) {
            fprintf(stderr,"Unexpected line at %d\n",linecount);
            continue;
         }
			iv1--; iv2--; iv3--;
			in1--; in2--; in3--;
			if (iv1 < 0 || iv1 >= nvertex) {
				fprintf(stderr,"Out of range vertex index (%d/%d) at %d\n",iv1,nvertex,linecount);
				continue;
			}
         if (iv2 < 0 || iv2 >= nvertex) {
            fprintf(stderr,"Out of range vertex index (%d/%d) at %d\n",iv2,nvertex,linecount);
            continue;
         }
         if (iv3 < 0 || iv3 >= nvertex) {
            fprintf(stderr,"Out of range vertex index (%d/%d) at %d\n",iv3,nvertex,linecount);
            continue;
         }
			polygons[npolygon].p1 = iv1;
         polygons[npolygon].p2 = iv2;
         polygons[npolygon].p3 = iv3;
         if (in1 < 0 || in1 >= nnormal) {
            fprintf(stderr,"Out of range normal index (%d/%d) at %d\n",iv1,nnormal,linecount);
            continue;
         }
         if (in2 < 0 || in2 >= nnormal) {
            fprintf(stderr,"Out of range normal index (%d/%d) at %d\n",iv2,nnormal,linecount);
            continue;
         }
         if (in3 < 0 || in3 >= nnormal) {
            fprintf(stderr,"Out of range normal index (%d/%d) at %d\n",iv3,nnormal,linecount);
            continue;
         }
         polygons[npolygon].n1 = in1;
         polygons[npolygon].n2 = in2;
         polygons[npolygon].n3 = in3;
			polygons[npolygon].group = igroup;
			npolygon++;
      }
	}

	return(TRUE);
}

void CleanExit(void)
{
	if (usewitilt) {
   	sleep(1);
   	SendSerialString(fdes," ");
		sleep(1);
		close(fdes);
	}
	exit(0);
}

void AddWitiltPoint(XYZ p)
{
	static int first = 0;
	static int iptr = 0;
	int i;
	static XYZ origin = {0,0,0};
	double scale = 0.25;

	if (first == 0) {
		for (i=0;i<AVERAGELEN;i++) 
			witilt_p[i] = origin;
	}

	if (first == 200)
		origin = p;

	witilt_p[iptr].x = (p.x - origin.x) * scale;
   witilt_p[iptr].y = (p.y - origin.y) * scale;
   witilt_p[iptr].z = (p.z - origin.z) * scale;

	iptr++;
	iptr %= AVERAGELEN;

	first++;
}

XYZ GetWitiltPoint(void)
{
	int i;
	XYZ sum = {0,0,0};

	for (i=0;i<AVERAGELEN;i++) {
		sum.x += witilt_p[i].x;
      sum.y += witilt_p[i].y;
      sum.z += witilt_p[i].z;
	}
	sum.x /= AVERAGELEN;
   sum.y /= AVERAGELEN;
   sum.z /= AVERAGELEN;

	return(sum);
}


